(function( $, document, params, accounting, undefined ) {
    "use strict";

	let VariationPrices = ( () => {

		const getPriceHTML = ( price ) => {
			const formatSettings = params.format_settings

			return params.price_template.replace( '%1$s', accounting.format( price, formatSettings.decimals, formatSettings.thousand_separator, formatSettings.decimal_separator ) )
		}

		const getSalePriceHTML = ( price, salePrice ) => {
			return params.single_format.replace( '%1$s', params.sale_price_template.replace( '%1$s', getPriceHTML( price ) ).replace( '%2$s', getPriceHTML( salePrice ) ) )
		}

		const getRangeHTML = ( min, max ) => {
			return params.range_format
						.replace( '%1$s', getPriceHTML( min ) )
						.replace( '%2$s', getPriceHTML( max ) ) + params.suffix_format
		}

		const getSingleHTML = ( price, salePrice = false ) => {
			return params.single_format.replace( '%1$s', getPriceHTML( price ) ) + params.suffix_format
		}

		const getVariationPriceHTML = ( variation ) => {
			const basePrice    = variation.adjustedPrice || variation.display_price,
				onSale       = variation.display_price !== variation.display_regular_price,
				regularPrice = variation.display_regular_price + variation.adjustedPrice - variation.display_price,
				singlePrice  = onSale ? getSalePriceHTML( regularPrice, basePrice ) : getSingleHTML( basePrice );

			return variation.maxAdjustment > 0 ? getRangeHTML( basePrice + variation.minAdjustment, basePrice + variation.maxAdjustment ) : singlePrice;
		}

		const getListItemHTML = ( variation, unchosenAttributes ) => {
			const name  = unchosenAttributes.map( k => decodeURIComponent( variation.attribute_names[k] ) ).filter( a => a ).join(', '),
				price = getVariationPriceHTML( variation )


			return params.list_item_template.replace( '%1$s', variation.variation_id )
											.replace( '%2$s', name )
											.replace( '%3$s', price )
											.trim();
		}

		const getListHTML = ( $form ) => {
			const formAttributes     = getFormAttributes( $form ),
				selectedVariations = getSelectedVariations( $form );

			const list = selectedVariations.variations.map( v => {
				return getListItemHTML( adjustVariation( v, $form ), formAttributes.unselected )
			}).join('');

			return params.list_template.replace( '%1$s', list )
		}

		const getSelectedVariations = ( $form ) => {
			const formAttributes     = getFormAttributes( $form ),
				productVariations  = $form.data('product_variations').sort( ( a, b ) => a.display_price < b.display_price ? 1 : -1 ),
				selectedVariations = productVariations.filter( pv => Object.entries( formAttributes.selected ).reduce( (s, e) => s + 1 * ( e[1] === pv.attributes[e[0]] || '' === pv.attributes[e[0]] ), 0 ) === Object.values( formAttributes.selected ).length ),
				min = selectedVariations.reduce( (min, v) => {
					const adjustedVariation = adjustVariation( v, $form )
					return Math.min( min, adjustedVariation.adjustedPrice + ( adjustedVariation.minAdjustment && adjustedVariation.minAdjustment < Infinity ? adjustedVariation.minAdjustment : 0 ) )
				}, Infinity ),
				max = selectedVariations.reduce( (max, v) => {
					const adjustedVariation = adjustVariation( v, $form )
					return Math.max( max, adjustedVariation.adjustedPrice + ( adjustedVariation.maxAdjustment || 0 ) )
				}, 0 );

			return {
				length: selectedVariations.length,
				variations: selectedVariations.sort( ( a, b ) => ( a.adjustedPrice + a.minAdjustment ) < ( b.adjustedPrice + b.minAdjustment ) ? 1 : -1 ),
				min,
				max
			}
		}

		const getFormAttributes = ( $form ) => {
			const selected = {},
				unselected = [];

			$form.find('select').each( (i, e) => {
				if ( $(e).val() ) {
					selected[ $(e).data('attribute_name') ] = $(e).val();
				} else {
					unselected.push( $(e).data('attribute_name') );
				}
			});

			return {
				selected,
				unselected
			}
		}

		const getDefaultHTML = ( $form ) => {
			const type = params.format_type;

			if ( 'list' === type ) {
				return getListHTML( $form );
			}

			const selectedVariations = getSelectedVariations( $form );

			return selectedVariations.min === selectedVariations.max ? getSingleHTML( selectedVariations.min ) : getRangeHTML( selectedVariations.min, selectedVariations.max );
		}

		const getVariationAdjustments = ( event, data ) => {
			const $form = $( event.target );

			if ( $form.find( '.iconic-was-fees' ).length ) {
				return JSON.parse( $form.find( '.iconic-was-fees' ).html() || data );
			}

			return data;
		}

		const adjustVariation = ( variation, $form ) => {
			let adjustments = {}
			const data = $form.triggerHandler( 'wvpAdjustments', [ adjustments ] );

			if ( undefined !== data ) {
				adjustments = data
			}

			const formAttributes = getFormAttributes( $form )
			const selectedAttributes = Object.fromEntries( Object.entries( formAttributes.selected ).map( e => [ e[0].replace('attribute_',''), e[1] ] ) );
			const attributes = Object.fromEntries( Object.entries( variation.attributes ).map( e => [ e[0].replace('attribute_',''), e[1] ] ) )
			let min = 0, max = 0;
			
			variation.adjustedPrice = variation.display_price

			for ( const attribute in attributes ) {
				let localMin = Infinity, localMax = 0;
				const currentAttribute = attributes[ attribute ] || selectedAttributes[ attribute ]

				if ( currentAttribute ) {
					if ( adjustments[ attribute ] && adjustments[ attribute ][ currentAttribute ] ) {
						variation.adjustedPrice += adjustments[ attribute ][ currentAttribute ].default;
					}
				} else {
					for ( const term in adjustments[ attribute ] ) {
						localMin = Math.min( localMin, adjustments[ attribute ][ term ].default );
						localMax = Math.max( localMax, adjustments[ attribute ][ term ].default );
					}
					min += localMin;
					max += localMax;
				}
			}
			if ( max > 0 ) {
				variation.minAdjustment = min
				variation.maxAdjustment = max
			} else {
				delete variation.minAdjustment
				delete variation.maxAdjustment
			}

			return variation
		}

		const updatePriceRange = ( $form ) => {
			if ( ! params.show_selected || $( 'select', $form ).length === 0 ) {
				return false;
			}

			const $wvp = $form.closest('.product[id^="product"], .product.product-row')
							.find( '.wvp-price-range' )
							// the script must only target the single product page price range
							// so we filter out all the price range elements that are inside a WC loop container (e.g. related products)
							.filter( ( index, element ) => 0 === $( element ).closest( '.products' ).length ); 

			if ( ! $form.data('product_variations') ) {
				// $wvp.html( getDefaultHTML( $form ) )
				// .toggleClass( 'price', ! $wvp.find( '.price' ).length );

				return false;
			}

			const formAttributes = getFormAttributes( $form );

			if ( formAttributes.unselected.length === $form.find('select').length ) {
				$wvp.html( getDefaultHTML( $form ) )
				.toggleClass( 'price', ! $wvp.find( '.price' ).length );

				return false;
			}

			const selectedVariations = getSelectedVariations( $form );

			if ( 'list' === params.format_type ) {

				// when the list of variation is selected, update the price only when all the attributes have been defined
				if ( ! params.show_selected || params.hide && selectedVariations.length > 1 ) {
					return false;
				}

				$wvp.html( getListHTML( $form ) )
					.toggleClass( 'price', ! $wvp.find( '.price' ).length );
			// } else if ( selectedVariations.length === 1 || selectedVariations.min === selectedVariations.max ) {
			// 	$wvp.html( getPriceHTML( selectedVariations.minVariation.display_price ) )
			// 		.toggleClass( 'price', ! $wvp.find( '.price' ).length );
			} else {
				$wvp.html( selectedVariations.min === selectedVariations.max ? getSingleHTML( selectedVariations.min ) : getRangeHTML( selectedVariations.min, selectedVariations.max ) )
					.toggleClass( 'price', ! $wvp.find( '.price' ).length );

				$wvp.find('del').remove();

				if ( $wvp.data('suffix') ) {
					// Fix the price suffix duplication
					$wvp.find( '.woocommerce-price-suffix' ).slice(0,-1).remove();

					const $suffix = $wvp.find( '.woocommerce-price-suffix' ).detach();
					$wvp.append( $suffix )
						.toggleClass( 'price', ! $wvp.find( '.price' ).length );
				} else {
					$wvp.find( '.woocommerce-price-suffix' ).remove();
				}
			}
		};

		const initialize = () => {
			// All the variations are listed as a JSON object
			// in the data-product_variations attribute
			$('[data-product_variations]').each( (i, e) => {
				const $form = $( e );
	
				$form.on( 'wvpAdjustments', getVariationAdjustments )
				updatePriceRange( $form );
			});
	
			// The number of variations exceeds the variation threshold
			// and the selected variation is loaded via AJAX
			// only after all attributes have been selected
			$( document ).on( 'found_variation', (event, variation) => {
				if ( ! params.show_selected ) {
					return false;
				}
	
				const $form  = $( event.target ),
					  $wvp   = $( event.target ).closest('.product').find( '.wvp-price-range' ).filter( (index, element) => 0 === $( element ).closest( '.products' ).length ),
					  $price = $( getVariationPriceHTML( adjustVariation( variation, $form ) ) );
	
				if ( 'list' !== params.format_type ) {
					$price.find('.woocommerce-price-suffix').remove();
				}
	
				$wvp.empty()
					.append( $price )
					.toggleClass( 'price', ! $wvp.find( '.price' ).length );
	
			});
	
			$('[data-product_variations] select').on('change', (event) => {
				const $form = $( event.target ).closest('form');
	
				updatePriceRange( $form );
			});
		}

		return {
			initialize,
			updatePriceRange,
			getPriceHTML,
			getSalePriceHTML,
			getRangeHTML,
			getSingleHTML
		}
	})();

	window.WCVariationPrices = VariationPrices;

	$( () => {
		if ( params.product_types.includes( 'variable' ) ) {
			VariationPrices.initialize();
		}
	});

})( jQuery, document, wcvp_params, accounting );
